# -*- coding: utf-8 -*-
import json
import typing
import uuid
from collections import namedtuple
from datetime import datetime
from urllib.parse import urlencode


from raven.contrib.tornado import SentryMixin

import tornado
from tornado.gen import coroutine
from tornado.httpclient import AsyncHTTPClient
from tornado.log import app_log
from tornado.web import RequestHandler

from .FourOhFour import FourOhFour
from .Multipart import Multipart

File = namedtuple('File', ['name', 'body', 'content_type', 'upload_name'])
CLOUD_EVENTS_FILE_KEY = '_ce_payload'


class ExecHandler(SentryMixin, RequestHandler):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.buffer = bytearray()
        self.response_passthrough = True
        self.response_multipart = None

    def prepare(self):
        self.set_header('Server', 'Storyscript')

    def resolve_by_uri(self, path):
        """
        A http request to `/*` will resolve to one listener on that channel.
        """
        resolve = self.application.router.find_handler(self.request)

        app_log.info(f'Resolving to {repr(resolve)}')

        if not resolve:
            # exit: path not being followed
            return None, None

        # Do not change the eventType and source.
        # The runtime makes the headers case-insensitive
        # if the eventType is http_request.
        # See
        # https://github.com/storyscript/platform-engine/pull/320/commits/cdb1df7f4fd36783c6b3b1d8ed92f539b672e388
        event = {
            'eventType': 'http_request',
            'cloudEventsVersion': '0.1',
            'source': 'gateway',
            'eventID': str(uuid.uuid4()),
            'eventTime': datetime.utcnow().replace(microsecond=0).isoformat(),
            'contentType': 'application/vnd.omg.object+json',
            'data': {
                'uri': self.request.uri,
                'path': self.request.path,
                'headers': dict(self.request.headers),
            },
        }

        # TODO: deprecate query_params once OMS supports deprecation
        # https://github.com/microservices/openmicroservices.org/pull/142
        event['data']['query_params'] = {}
        for k, v in self.request.arguments.items():
            event['data']['query_params'][k] = v[0].decode('utf-8')

        event['data']['queryParams'] = {}
        for k, v in self.request.query_arguments.items():
            event['data']['queryParams'][k] = v[0].decode('utf-8')

        event['data']['formFields'] = {}
        for k, v in self.request.body_arguments.items():
            event['data']['formFields'][k] = v[0].decode('utf-8')

        if 'application/json' in self.request.headers.get('content-type', ''):
            event['data']['body'] = json.loads(
                self.request.body.decode('utf-8')
            )

        return resolve, event

    @coroutine
    def _handle(self, path):
        resolve, event = self.resolve_by_uri(path)

        if resolve is None and event is None:
            FourOhFour.handle(self)
            self.finish()
            return

        url = resolve.endpoint

        try:
            yield self.execute_request(url, event)
        except:
            import traceback

            traceback.print_exc()
            self.set_status(500, reason='Story execution failed')
            self.write('HTTP 500: Story execution failed\n')

        if not self._finished:
            self.finish()

    @coroutine
    def execute_request(self, url, event):
        """
        If there are any files in the request, then the request made
        to the engine will be multipart/form-data.
        If no files exist, then it'll be a plain old application/json request.
        """
        kwargs = {
            'method': 'POST',
            'url': url,
            'connect_timeout': 10,
            'request_timeout': 60,
            'streaming_callback': self._callback,
            'header_callback': self._on_headers_receive,
        }

        if len(self.request.files) == 0:
            kwargs['body'] = json.dumps(event)
            kwargs['headers'] = {
                'Content-Type': 'application/json; charset=utf-8'
            }
        else:
            boundary = uuid.uuid4().hex
            headers = {
                'Content-Type': f'multipart/form-data; boundary={boundary}'
            }
            files = self._get_request_files()
            self._insert_event_as_file(event, files)
            kwargs['headers'] = headers
            # body_producer is NOT supported for the curl_httpclient
            # https://github.com/tornadoweb/tornado/issues/2096
            kwargs['body'] = b''.join(self.multipart_producer(files, boundary))

        kwargs['headers']['Connection'] = 'keep-alive'
        request = tornado.httpclient.HTTPRequest(**kwargs)
        client = AsyncHTTPClient()
        yield client.fetch(request)
        # make sure that the response buffer gets flushed
        self._callback(b'', final=True)

    def _insert_event_as_file(self, event: dict, files: typing.List[File]):
        files.append(
            File(
                name=CLOUD_EVENTS_FILE_KEY,
                body=json.dumps(event).encode('utf-8'),
                content_type='application/json',
                upload_name=CLOUD_EVENTS_FILE_KEY,
            )
        )

    def _get_request_files(self) -> typing.List[File]:
        # File handling:
        # self.request.files looks like this:
        # {"upload_name": [{filename:<>, body:<>, content_type:<>}, {<file>}]}
        all_files = []
        for upload_name in self.request.files:
            for _f in self.request.files[upload_name]:
                all_files.append(
                    File(
                        name=_f['filename'],
                        body=_f['body'],
                        content_type=_f['content_type'],
                        upload_name=upload_name,
                    )
                )
        return all_files

    def multipart_producer(self, files: typing.List[File], boundary):
        """
        Inspired directly from here:
        https://github.com/tornadoweb/tornado/blob/master/demos/file_upload/file_uploader.py
        """
        boundary_bytes = boundary.encode()

        for file in files:
            filename_bytes = file.name.encode()
            upload_name_bytes = file.upload_name.encode()
            buf = (
                (b'--%s\r\n' % boundary_bytes)
                + (
                    b'Content-Disposition: form-data; '
                    b'name="%s"; filename="%s"\r\n'
                    % (upload_name_bytes, filename_bytes)
                )
                + (b'Content-Type: %s\r\n' % file.content_type.encode())
                + b'\r\n'
            )
            yield buf

            # We only write bytes.
            assert isinstance(file.body, bytes)
            yield file.body

            yield b'\r\n'

        yield b'--%s--\r\n' % (boundary_bytes,)

    def _on_headers_receive(self, header_line: str):
        """
        Checks if a Content-Type header is sent, which indicates if the data
        received is binary. If it's binary, it needs to be streamed to the
        client without the usual command processing logic.
        """
        if header_line.lower().startswith('content-type'):
            parts = header_line.split(':')
            value = parts[1].strip()
            if value.startswith('application/stream+json'):
                self.response_passthrough = False
            elif value.startswith('multipart/mixed'):
                idx = value.find("boundary=")
                assert idx >= 0, f'No boundary in {value}'
                boundary = value[idx + 9:]
                response_boundary = boundary.partition(";")[0]
                if response_boundary.startswith('"'):
                    assert response_boundary.endswith('"')
                    response_boundary = response_boundary[1:-1]
                self.response_multipart = Multipart(response_boundary,
                                                    self.process_chunk)
            else:
                # Since it's not json, push the response to the client as is.
                self.set_header('Content-Type', value)

    def _callback(self, chunk, final=False):
        """
        Chunk examples that come from the Engine
            set_status 200
            set_header {"name":"X-Data", "value":"Storyscript"}
            write Hello, world
            ~finish~ will not be passed since it will close the connection
        """

        if self.response_multipart is not None:
            self.response_multipart.process(chunk, final)
            return
        # If the response from the engine is binary (see _on_headers_receive),
        # then the response must be sent to the client directly.
        elif self.response_passthrough:
            self.write(chunk)
            return

        self.process_json_chunk(chunk)

    def process_chunk(self, chunk, content_type):
        if content_type == 'application/stream+json':
            self.process_json_chunk(chunk)
            return
        else:
            if len(chunk) > 0:
                self.write(chunk)
            return

    def process_json_chunk(self, chunk):
        # Read `chunk` byte by byte and add it to the buffer.
        # When a byte is \n, then parse everything in the buffer as string,
        # and interpret the resulting JSON string.

        instructions = []
        for b in chunk:
            if b == 0x0A:  # 0x0A is an ASCII/UTF-8 new line.
                instructions.append(self.buffer.decode('utf-8'))
                self.buffer.clear()
            else:
                self.buffer.append(b)

        # If we have any new instructions, execute them.
        for ins in instructions:
            if ins.isspace():
                continue
            ins = json.loads(ins)
            command = ins['command']
            if command == 'write':
                if ins['data'].get('content') is None:
                    self.write('null')
                else:
                    self.write(ins['data']['content'])
                if ins['data'].get('flush'):
                    self.flush()
            elif command == 'writeJSON':
                self.set_header('Content-Type',
                                'application/json; charset=utf-8')
                self.write(json.dumps(ins['data']['content']))
                self.flush()
            elif command == 'set_status':
                self.set_status(ins['data']['code'])
            elif command == 'set_cookie':
                # name, value, domain, expires, path, expires_days, secure
                if ins['data'].pop('secure', False):
                    self.set_cookie(**ins['data'])
                else:
                    self.set_secure_cookie(**ins['data'])
            elif command == 'clear_cookie':
                # name, domain, path
                self.clear_cookie(**ins['data'])
            elif command == 'clear_all_cookie':
                # domain, path
                self.clear_cookie(**ins['data'])
            elif command == 'set_header':
                self.set_header(ins['data']['key'], ins['data']['value'])
            elif command == 'flush':
                self.flush()
            elif command == 'redirect':
                redir_url = ins['data']['url']
                params = ins['data'].get('query')
                if isinstance(params, dict):
                    # Convert boolean True/False to true/false as strings,
                    # because of Python's silly-ness.
                    self.handle_boolean_values(params)
                    query_string = urlencode(params)
                    if '?' in redir_url:
                        redir_url = f'{redir_url}&{query_string}'
                    else:
                        redir_url = f'{redir_url}?{query_string}'
                self.redirect(redir_url)
            elif command == 'finish':
                # can we close quicker here?
                break
            else:
                raise NotImplementedError(f'{command} is not implemented!')

    def handle_boolean_values(self, params: dict):
        for k, v in params.items():
            if isinstance(v, bool):
                if v:
                    params[k] = 'true'
                else:
                    params[k] = 'false'

    @coroutine
    def head(self, path):
        yield self._handle(path)

    @coroutine
    def get(self, path):
        yield self._handle(path)

    @coroutine
    def post(self, path):
        yield self._handle(path)

    @coroutine
    def delete(self, path):
        yield self._handle(path)

    @coroutine
    def patch(self, path):
        yield self._handle(path)

    @coroutine
    def put(self, path):
        yield self._handle(path)

    def options(self, path):
        """
        Returns the allowed options for this endpoint
        """
        self.set_header('Allow', 'GET,HEAD,POST,PUT,PATCH,DELETE,OPTIONS')
        # [FUTURE] http://zacstewart.com/2012/04/14/http-options-method.html
        self.finish()
