from enum import Enum


class MultipartState(Enum):
    """
    Different states of the multipart response.
    """
    NO_MULTIPART = 0
    WAIT_FOR_BOUNDARY = 1
    PARSE_HEADERS = 2
    FINISHED = 3


class Multipart:
    """
    Read an incoming multipart response split into chunks.
    """
    def __init__(self, boundary, processor):
        self.buffer = bytearray()
        self.boundary = f'\r\n--{boundary}'.encode()
        self.state = MultipartState.WAIT_FOR_BOUNDARY
        # + 2 for the final '--' or CRLF
        self.lookahead = len(self.boundary) * 2 + 2
        self.content_type = None
        self._processor = processor

    def process(self, chunk, final):
        if self.state == MultipartState.FINISHED:
            # final boundary was reached - no further processing
            return

        if self.state == MultipartState.WAIT_FOR_BOUNDARY:
            self.wait_for_boundary(chunk, final)
            return

        assert self.state == MultipartState.PARSE_HEADERS, \
            f'Unknown state: {self.state}'

        self.parse_headers(chunk, final)

    def switch_state(self, chunk, final, state):
        self.state = state
        self.buffer.clear()
        self.process(chunk, final)

    def parse_headers(self, chunk, final):
        """
        Parses a 'Content-Type' line from the stream.
        """
        self.buffer.extend(chunk)
        if self.buffer.find(b'\n') == -1:
            return

        # parse the new contentType
        line, _, other = self.buffer.partition(b'\n')
        # in contrast to RFC1341 we always require a content-type
        line = line.decode('utf-8')
        assert line.lower().startswith('content-type:')
        self.content_type = line.split(':')[1].strip()

        self.switch_state(other, final, MultipartState.WAIT_FOR_BOUNDARY)

    def ensure_buffer_lookahead(self, chunk, final):
        """
        Wait until before has reached lookahead size
        """
        # buffer must be at least bigger than the boundary
        if len(chunk) < self.lookahead:
            if len(self.buffer) < self.lookahead and not final:
                # buffer is still to small
                self.buffer.extend(chunk)
                return None
            chunk = self.buffer + chunk
        else:
            # check for previous chunks
            if len(self.buffer) > 0:
                chunk = self.buffer + chunk

        self.buffer.clear()
        return chunk

    def wait_for_boundary(self, chunk, final):
        """
        Parses the chunk stream for the multipart boundary.
        """
        chunk = self.ensure_buffer_lookahead(chunk, final)
        if chunk is None:
            return

        # see https://www.w3.org/Protocols/rfc1341/7_2_Multipart.html
        # detect starting boundary
        before, sep, after = chunk.partition(self.boundary)

        # check if boundary was found
        if len(sep) == 0:
            # no boundary -> print all
            self.process_chunk(chunk)
            return

        # print previous parts before content type switch
        self.process_chunk(before)

        # check for final boundary -> no further processing
        if after.startswith(b'--'):
            self.state = MultipartState.FINISHED
            return

        # boundary must be followed immediately by another CRLF and headers
        assert after.startswith(b'\r\n'), \
            'Boundary must be followed by either "--" (final) or CRLF'
        after = after[2:]

        self.switch_state(after, final, MultipartState.PARSE_HEADERS)

    def process_chunk(self, chunk):
        """
        Forwards the chunk to the underlying readers based on the contentType.
        """
        if len(chunk) == 0:
            return
        if self.content_type is not None:
            self._processor(chunk, self.content_type)
