# -*- coding: utf-8 -*-
from unittest import mock

from pytest import fixture, raises

from app.handlers.Exec import ExecHandler


@fixture
def handler(mocker):
    magic = mocker.MagicMock
    _handler = ExecHandler(application=magic(), request=magic())
    mocker.patch.object(_handler, 'flush')
    mocker.patch.object(_handler, 'set_header')

    _handler.result = bytearray()

    def concat_write(chunk):
        _handler.result.extend(chunk)

    mocker.patch.object(_handler, 'write', side_effect=concat_write)
    return _handler


@fixture
def hs(handler):
    header = 'Content-Type: multipart/mixed; boundary=foo'
    handler._on_headers_receive(header)
    return handler


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i + n]


def build_chunk(boundary, content_type, chunk):
    return (f'\r\n--{boundary}\r\nContent-Type: {content_type}'
            f'\n{chunk}'.encode())


def test_multipart_header(handler):
    header = 'Content-Type: multipart/mixed;boundary=foo'
    handler._on_headers_receive(header)
    assert handler.response_multipart.boundary == '\r\n--foo'.encode()
    handler.set_header.assert_not_called()


def test_multipart_header_throws(handler):
    header = 'Content-Type: multipart/mixed'
    with raises(AssertionError):
        handler._on_headers_receive(header)


def test_multipart_header_quoted(handler):
    header = 'Content-Type: multipart/mixed; boundary="barI78$"'
    handler._on_headers_receive(header)
    assert handler.response_multipart.boundary == '\r\n--barI78$'.encode()
    handler.set_header.assert_not_called()


def test_multipart_single(hs):
    hs._callback(build_chunk(
        'foo', 'application/stream+json',
        '{"command": "set_header", "data": '
        '{"key": "Content-Type", "value": "text/plain"}}\n'))

    hs.set_header.assert_called_with('Content-Type', 'text/plain')
    hs.write.assert_not_called()


def test_multipart_single_boundary(hs):
    hs._callback(build_chunk('foo', 'application/octet-stream', '--bar'),
                 final=True)
    hs.result = b'--bar'
    hs.set_header.assert_not_called()


def test_multipart_invalid_boundary(hs):
    with raises(AssertionError):
        hs._callback(b'\r\n--fooa', final=True)


def test_multipart_multiple_chunks(hs):
    hs._callback(build_chunk('foo', 'application/octet-stream', 'chunk1'))
    hs._callback(b'chunk2', final=True)
    hs.result == b'chunk1chunk2'
    hs.set_header.assert_not_called()


def test_multipart_multiple_chunks_long(hs):
    hs._callback(build_chunk('foo', 'application/octet-stream',
                 'chunk1.chunk2.chunk3.chunk4.chunk5.chunk6'))
    hs._callback(b'chunk1b.chunk2b.chunk3b.chunk4b.chunk5b.chunk6b')

    assert hs.result == (
        b'chunk1.chunk2.chunk3.chunk4.chunk5.chunk6'
        b'chunk1b.chunk2b.chunk3b.chunk4b.chunk5b.chunk6b'
    )
    hs.set_header.assert_not_called()


def test_multipart_mixed(hs):
    hs._callback(build_chunk(
        'foo', 'application/stream+json',
        '{"command": "set_header", "data": '
        '{"key": "Content-Type", "value": "text/plain"}}\n'))

    hs.set_header.assert_called_with('Content-Type', 'text/plain')
    hs.write.assert_not_called()
    hs.set_header.reset_mock()

    hs._callback(build_chunk('foo', 'application/octet-stream',
                 'data.chunk1.chunk2.chunk3'))
    hs.result == b'data.chunk1.chunk2.chunk3'
    hs.set_header.assert_not_called()


def test_multipart_mixed_three(hs):
    hs._callback(build_chunk(
        'foo', 'application/stream+json',
        '{"command": "set_header", "data": '
        '{"key": "Content-Type", "value": "text/plain"}}\n'))

    hs.set_header.assert_called_with('Content-Type', 'text/plain')
    hs.set_header.reset_mock()
    hs.write.assert_not_called()

    hs._callback(build_chunk('foo', 'application/octet-stream',
                 'data.chunk1.chunk2.chunk3'))
    hs.set_header.assert_not_called()
    hs.result == b'data.chunk1.chunk2.chunk3'
    hs.write.reset_mock()

    hs._callback(build_chunk(
        'foo', 'application/stream+json', '{"command": "flush"}\n'))
    hs.set_header.assert_not_called()
    hs.write.assert_not_called()
    hs.flush.assert_called()


def test_multipart_small_boundary(handler):
    boundary = 'boundary01234567890123456789'
    header = f'Content-Type: multipart/mixed; boundary={boundary}'
    handler._on_headers_receive(header)
    hs = handler

    hs._callback(build_chunk(boundary, 'application/octet-stream', 'chunk'))
    for i in range(20):
        hs._callback(f'.chunk{i}'.encode())

    hs._callback(b'', final=True)
    hs.result = (
        b'chunk.chunk0.chunk1.chunk2.chunk3.chunk4'
        b'.chunk5.chunk6.chunk7.chunk8'
        b'.chunk9.chunk10.chunk11.chunk12.chunk13'
        b'.chunk14.chunk15.chunk16.chunk17'
        b'.chunk18.chunk19'
    )
    hs.set_header.assert_not_called()


def test_multipart_small_boundary2(handler):
    boundary = 'boundary01234567890123456789'
    header = f'Content-Type: multipart/mixed; boundary={boundary}'
    handler._on_headers_receive(header)
    hs = handler

    hs._callback(build_chunk(boundary, 'application/stream+json', '{'))
    hs._callback(b'"command":')
    hs._callback(b' "flush"}')
    hs._callback(b'\n')

    for chunk in chunks(build_chunk(boundary, 'application/octet-stream', 'chunk1'), 4):
        hs._callback(chunk)
    hs._callback(b'', final=True)

    assert hs.result == b'chunk1'
    hs.set_header.assert_not_called()
    hs.flush.assert_called()
